from OgreInterface.score_function.ewald import EnergyEwald
from OgreInterface.score_function.born import EnergyBorn
from OgreInterface.score_function.generate_inputs import generate_dict_torch
from OgreInterface.surfaces import Interface
from pymatgen.io.ase import AseAtomsAdaptor
from pymatgen.core.periodic_table import Element
from pymatgen.analysis.local_env import CrystalNN
from ase.data import atomic_numbers, chemical_symbols
from typing import Dict
import numpy as np
from matplotlib.colors import Normalize
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
from scipy.interpolate import RectBivariateSpline, CubicSpline
from copy import deepcopy
from itertools import groupby, combinations_with_replacement, product


class IonicSurfaceMatcher:
    def __init__(
        self,
        interface: Interface,
        grid_density_x: int = 15,
        grid_density_y: int = 15,
    ):
        self.interface = interface
        self.matrix = deepcopy(interface.interface.lattice.matrix)
        self._vol = np.linalg.det(self.matrix)

        if self._vol < 0:
            self.matrix *= -1
            self._vol *= -1

        self.cutoff, self.alpha, self.k_max = self._get_ewald_parameters()
        self.k_max = 10
        self.charge_dict = self._get_charges()
        self.r0_dict = self._get_r0s(
            sub=self.interface.substrate.conventional_bulk_structure,
            film=self.interface.film.conventional_bulk_structure,
            charge_dict=self.charge_dict,
        )
        self.ns_dict = {element: 6.0 for element in self.charge_dict}
        self.d_interface = self.interface.interfacial_distance
        self.film_part = self.interface.film_part
        self.sub_part = self.interface.sub_part
        self.grid_density_x = grid_density_x
        self.grid_density_y = grid_density_y
        self.opt_xy_shift = np.zeros(3)
        self.opt_z_shift = np.zeros(3)

        self.shifts, self.X, self.Y = self._generate_shifts()
        self.z_PES_data = None

    def get_optmized_structure(self):
        opt_shift = self.opt_xy_shift + self.opt_z_shift
        self.interface.shift_film(
            shift=opt_shift, fractional=True, inplace=True
        )

    # def _get_charges_and_radii(self):
    #     sub = self.interface.substrate.conventional_bulk_structure
    #     film = self.interface.film.conventional_bulk_structure
    #     sub_oxidation_state = sub.composition.oxi_state_guesses()[0]
    #     film_oxidation_state = film.composition.oxi_state_guesses()[0]

    #     sub_oxidation_state.update(film_oxidation_state)

    #     radii = {
    #         element: Element(element).ionic_radii[charge]
    #         for element, charge in sub_oxidation_state.items()
    #     }

    #     return sub_oxidation_state, radii

    def _get_charges(self):
        sub = self.interface.substrate.conventional_bulk_structure
        film = self.interface.film.conventional_bulk_structure
        sub_oxidation_state = sub.composition.oxi_state_guesses()[0]
        film_oxidation_state = film.composition.oxi_state_guesses()[0]

        sub_oxidation_state.update(film_oxidation_state)

        return sub_oxidation_state

    def _get_neighborhood_info(self, struc, charge_dict):
        struc.add_oxidation_state_by_element(charge_dict)
        Zs = np.unique(struc.atomic_numbers)
        combos = combinations_with_replacement(Zs, 2)
        neighbor_dict = {c: None for c in combos}

        neighbor_list = []

        cnn = CrystalNN()
        for i, site in enumerate(struc.sites):
            info_dict = cnn.get_nn_info(struc, i)
            for neighbor in info_dict:
                dist = site.distance(neighbor["site"])
                species = tuple(
                    sorted([site.specie.Z, neighbor["site"].specie.Z])
                )
                neighbor_list.append([species, dist])

        sorted_neighbor_list = sorted(neighbor_list, key=lambda x: x[0])
        groups = groupby(sorted_neighbor_list, key=lambda x: x[0])

        for group in groups:
            nn = list(zip(*group[1]))[1]
            neighbor_dict[group[0]] = np.min(nn)

        for n, d in neighbor_dict.items():
            s1 = chemical_symbols[n[0]]
            s2 = chemical_symbols[n[1]]
            c1 = charge_dict[s1]
            c2 = charge_dict[s2]

            if d is None:
                d1_ionic = float(Element(s1).ionic_radii[c1])
                d2_ionic = float(Element(s2).ionic_radii[c2])

                neighbor_dict[n] = d1_ionic + d2_ionic

        return neighbor_dict

    def _get_r0s(self, sub, film, charge_dict):
        sub_dict = self._get_neighborhood_info(sub, charge_dict)
        film_dict = self._get_neighborhood_info(film, charge_dict)

        interface_atomic_numbers = np.unique(
            np.concatenate([sub.atomic_numbers, film.atomic_numbers])
        )

        ionic_radius_dict = {
            n: Element(chemical_symbols[n]).ionic_radii[
                charge_dict[chemical_symbols[n]]
            ]
            for n in interface_atomic_numbers
        }
        interface_combos = product(interface_atomic_numbers, repeat=2)
        interface_neighbor_dict = {}
        for c in interface_combos:
            interface_neighbor_dict[(0, 0) + c] = None
            interface_neighbor_dict[(1, 1) + c] = None
            interface_neighbor_dict[(0, 1) + c] = None
            interface_neighbor_dict[(1, 0) + c] = None

        all_keys = np.array(list(sub_dict.keys()) + list(film_dict.keys()))
        unique_keys = np.unique(all_keys, axis=0)
        unique_keys = list(map(tuple, unique_keys))

        for key in unique_keys:
            rev_key = tuple(reversed(key))
            sum_d = ionic_radius_dict[key[0]] + ionic_radius_dict[key[1]]
            if key in sub_dict and key in film_dict:
                sub_d = sub_dict[key]
                film_d = film_dict[key]
                interface_neighbor_dict[(0, 0) + key] = sub_d
                interface_neighbor_dict[(1, 1) + key] = film_d
                interface_neighbor_dict[(0, 1) + key] = (sub_d + film_d) / 2
                interface_neighbor_dict[(1, 0) + key] = (sub_d + film_d) / 2
                interface_neighbor_dict[(0, 0) + rev_key] = sub_d
                interface_neighbor_dict[(1, 1) + rev_key] = film_d
                interface_neighbor_dict[(0, 1) + rev_key] = (
                    sub_d + film_d
                ) / 2
                interface_neighbor_dict[(1, 0) + rev_key] = (
                    sub_d + film_d
                ) / 2

            if key in sub_dict and key not in film_dict:
                sub_d = sub_dict[key]
                interface_neighbor_dict[(0, 0) + key] = sub_d
                interface_neighbor_dict[(1, 1) + key] = sum_d
                interface_neighbor_dict[(0, 1) + key] = sub_d
                interface_neighbor_dict[(1, 0) + key] = sub_d
                interface_neighbor_dict[(0, 0) + rev_key] = sub_d
                interface_neighbor_dict[(1, 1) + rev_key] = sum_d
                interface_neighbor_dict[(0, 1) + rev_key] = sub_d
                interface_neighbor_dict[(1, 0) + rev_key] = sub_d

            if key not in sub_dict and key in film_dict:
                film_d = film_dict[key]
                interface_neighbor_dict[(1, 1) + key] = film_d
                interface_neighbor_dict[(0, 0) + key] = sum_d
                interface_neighbor_dict[(0, 1) + key] = film_d
                interface_neighbor_dict[(1, 0) + key] = film_d
                interface_neighbor_dict[(1, 1) + rev_key] = film_d
                interface_neighbor_dict[(0, 0) + rev_key] = sum_d
                interface_neighbor_dict[(0, 1) + rev_key] = film_d
                interface_neighbor_dict[(1, 0) + rev_key] = film_d

            if key not in sub_dict and key not in film_dict:
                interface_neighbor_dict[(0, 0) + key] = sub_d
                interface_neighbor_dict[(1, 1) + key] = sum_d
                interface_neighbor_dict[(0, 1) + key] = sum_d
                interface_neighbor_dict[(1, 0) + key] = sum_d
                interface_neighbor_dict[(0, 0) + rev_key] = sub_d
                interface_neighbor_dict[(1, 1) + rev_key] = sum_d
                interface_neighbor_dict[(0, 1) + rev_key] = sum_d
                interface_neighbor_dict[(1, 0) + rev_key] = sum_d

        for key, val in interface_neighbor_dict.items():
            if val is None:
                sum_d = ionic_radius_dict[key[2]] + ionic_radius_dict[key[3]]
                interface_neighbor_dict[key] = sub_d

        return interface_neighbor_dict

    def _get_ewald_parameters(self):
        struc_vol = self.interface._structure_volume
        # struc_vol = self._vol
        accf = np.sqrt(np.log(10**4))
        w = 1 / 2**0.5
        alpha = np.pi * (
            len(self.interface.interface) * w / (struc_vol**2)
        ) ** (1 / 3)
        cutoff = accf / np.sqrt(alpha)
        k_max = 2 * np.sqrt(alpha) * accf

        return cutoff, alpha, k_max

    def _generate_shifts(self):
        grid_x = np.linspace(0, 1, self.grid_density_x)
        grid_y = np.linspace(0, 1, self.grid_density_y)
        X, Y = np.meshgrid(grid_x, grid_y)

        shifts = np.c_[X.ravel(), Y.ravel(), np.zeros(X.shape).ravel()]

        return shifts, X, Y

    def _get_shifted_atoms_old(self, shifts):

        atoms = [
            AseAtomsAdaptor().get_atoms(self.sub_part),
            AseAtomsAdaptor().get_atoms(self.film_part),
        ]

        for shift in shifts:
            shifted_atoms = self.interface.shift_film(
                shift, fractional=True, inplace=False, return_atoms=True
            )
            atoms.append(shifted_atoms)

        return atoms

    def _get_shifted_atoms(self, shifts):
        atoms = []
        z_atoms = []

        z_shift = np.array([0, 0, 10]) / np.linalg.norm(self.matrix[-1])

        for shift in shifts:
            shifted_atoms = self.interface.shift_film(
                shift + self.opt_z_shift,
                fractional=True,
                inplace=False,
                return_atoms=True,
            )
            shifted_atoms.set_array(
                "is_film", self.interface.interface.site_properties["is_film"]
            )
            z_shifted_atoms = self.interface.shift_film(
                shift + z_shift,
                fractional=True,
                inplace=False,
                return_atoms=True,
            )
            z_shifted_atoms.set_array(
                "is_film", self.interface.interface.site_properties["is_film"]
            )
            atoms.append(shifted_atoms)
            z_atoms.append(z_shifted_atoms)

        return atoms, z_atoms

    def _get_shifted_atoms_z(self, shifts):
        z_shift = np.array([0, 0, 10]) / np.linalg.norm(self.matrix[-1])
        atoms = [
            self.interface.shift_film(
                z_shift + self.opt_xy_shift,
                fractional=True,
                inplace=False,
                return_atoms=True,
            )
        ]
        atoms[0].set_array(
            "is_film", self.interface.interface.site_properties["is_film"]
        )

        for shift in shifts:
            shifted_atoms = self.interface.shift_film(
                shift + self.opt_xy_shift,
                fractional=True,
                inplace=False,
                return_atoms=True,
            )
            shifted_atoms.set_array(
                "is_film", self.interface.interface.site_properties["is_film"]
            )
            atoms.append(shifted_atoms)

        return atoms

    def _get_shifted_strucs(self, shifts):

        atoms = [
            self.sub_part,
            self.film_part,
        ]

        for shift in shifts:
            shifted_atoms = self.interface.shift_film(
                shift, fractional=True, inplace=False, return_atoms=False
            )
            atoms.append(shifted_atoms)

        return atoms

    def _generate_inputs(self, atoms_list):
        inputs = generate_dict_torch(
            atoms=atoms_list,
            cutoff=self.cutoff,
            charge_dict=self.charge_dict,
            # radius_dict=self.radius_dict,
            ns_dict=self.ns_dict,
        )

        return inputs

    def _calculate_coulomb(self, inputs):
        ewald = EnergyEwald(alpha=self.alpha, k_max=self.k_max)
        coulomb_energy = ewald.forward(inputs)

        return coulomb_energy

    def _calculate_born(self, inputs):
        born = EnergyBorn(cutoff=self.cutoff)
        born_energy = born.forward(inputs, r0_dict=self.r0_dict)

        return born_energy

    def _get_interpolated_data(self, X, Y, Z):
        x_grid = np.linspace(0, 1, self.grid_density_x)
        y_grid = np.linspace(0, 1, self.grid_density_y)
        spline = RectBivariateSpline(y_grid, x_grid, Z)

        x_grid_interp = np.linspace(0, 1, 401)
        y_grid_interp = np.linspace(0, 1, 401)

        X_interp, Y_interp = np.meshgrid(x_grid_interp, y_grid_interp)
        Z_interp = spline.ev(xi=Y_interp, yi=X_interp)
        frac_shifts = np.c_[
            X_interp.ravel(),
            Y_interp.ravel(),
            np.zeros(X_interp.shape).ravel(),
        ]

        cart_shifts = frac_shifts.dot(self.matrix)

        X_cart = cart_shifts[:, 0].reshape(X_interp.shape)
        Y_cart = cart_shifts[:, 1].reshape(Y_interp.shape)

        return X_cart, Y_cart, Z_interp

    def _plot_heatmap(
        self, fig, ax, X, Y, Z, borders, cmap, fontsize, show_max
    ):
        ax.set_xlabel(r"Shift in $x$ ($\AA$)", fontsize=fontsize)
        ax.set_ylabel(r"Shift in $y$ ($\AA$)", fontsize=fontsize)

        im = ax.contourf(
            X,
            Y,
            Z,
            cmap=cmap,
            levels=200,
            norm=Normalize(vmin=np.nanmin(Z), vmax=np.nanmax(Z)),
        )

        ax.plot(
            borders[:, 0],
            borders[:, 1],
            color="black",
            linewidth=2,
        )

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("top", size="5%", pad=0.1)
        cbar = fig.colorbar(im, cax=cax, orientation="horizontal")
        cbar.ax.tick_params(labelsize=fontsize)
        cbar.ax.locator_params(nbins=3)

        if show_max:
            E_max = np.max(Z)
            label = (
                "$E_{adh}$ (eV/$\\AA^{2}$) : "
                + "$E_{max}$ = "
                + f"{E_max:.4f}"
            )
            cbar.set_label(label, fontsize=fontsize)
        else:
            label = "$E_{adh}$ (eV/$\\AA^{2}$)"
            cbar.set_label(label, fontsize=fontsize)

        cax.xaxis.set_ticks_position("top")
        cax.xaxis.set_label_position("top")
        ax.tick_params(labelsize=fontsize)
        ax.set_xlim(borders[:, 0].min(), borders[:, 0].max())
        ax.set_ylim(borders[:, 1].min(), borders[:, 1].max())
        ax.set_aspect("equal")

    def run_surface_matching(
        self,
        cmap: str = "jet",
        fontsize: int = 14,
        output: str = "PES.png",
        shift: bool = True,
        show_born_and_coulomb: bool = False,
        dpi: int = 400,
        show_max: bool = False,
    ) -> float:
        shifts = self.shifts
        atoms_list, z_atoms_list = self._get_shifted_atoms(shifts)

        inputs = self._generate_inputs(atoms_list)
        if self.z_PES_data is None:
            z_inputs = self._generate_inputs(z_atoms_list)
            z_coulomb_energy = self._calculate_coulomb(z_inputs)
            z_born_energy = self._calculate_born(z_inputs)
            z_interface_coulomb_energy = z_coulomb_energy.reshape(self.X.shape)
            z_interface_born_energy = z_born_energy.reshape(self.X.shape)
            self.z_PES_data = [
                z_interface_coulomb_energy,
                z_interface_born_energy,
            ]
        else:
            z_interface_coulomb_energy = self.z_PES_data[0]
            z_interface_born_energy = self.z_PES_data[1]

        coulomb_energy = self._calculate_coulomb(inputs)
        born_energy = self._calculate_born(inputs)

        interface_coulomb_energy = coulomb_energy.reshape(self.X.shape)
        interface_born_energy = born_energy.reshape(self.X.shape)

        coulomb_adh_energy = (
            z_interface_coulomb_energy - interface_coulomb_energy
        )
        born_adh_energy = z_interface_born_energy - interface_born_energy

        X_plot, Y_plot, Z_born = self._get_interpolated_data(
            self.X, self.Y, born_adh_energy
        )
        _, _, Z_coulomb = self._get_interpolated_data(
            self.X, self.Y, coulomb_adh_energy
        )

        Z_both = Z_born + Z_coulomb

        a = self.matrix[0, :2]
        b = self.matrix[1, :2]
        borders = np.vstack([np.zeros(2), a, a + b, b, np.zeros(2)])
        x_size = borders[:, 0].max() - borders[:, 0].min()
        y_size = borders[:, 1].max() - borders[:, 1].min()
        ratio = y_size / x_size

        if show_born_and_coulomb:
            fig, (ax1, ax2, ax3) = plt.subplots(
                figsize=(3 * 5, 5 * ratio),
                ncols=3,
                dpi=dpi,
            )

            self._plot_heatmap(
                fig=fig,
                ax=ax1,
                X=X_plot,
                Y=Y_plot,
                Z=Z_born / (self.interface.area),
                borders=borders,
                cmap=cmap,
                fontsize=fontsize,
                show_max=False,
            )
            self._plot_heatmap(
                fig=fig,
                ax=ax2,
                X=X_plot,
                Y=Y_plot,
                Z=Z_coulomb / (self.interface.area),
                borders=borders,
                cmap=cmap,
                fontsize=fontsize,
                show_max=False,
            )
            self._plot_heatmap(
                fig=fig,
                ax=ax3,
                X=X_plot,
                Y=Y_plot,
                Z=Z_both / (self.interface.area),
                borders=borders,
                cmap=cmap,
                fontsize=fontsize,
                show_max=show_max,
            )

            frac_shifts = np.c_[
                X_plot.ravel(), Y_plot.ravel(), np.zeros(Y_plot.shape).ravel()
            ].dot(np.linalg.inv(self.matrix))
            opt_shift = frac_shifts[np.argmax(Z_both.ravel())]
            max_Z = np.max(Z_both)
            plot_shift = opt_shift.dot(self.matrix)

            ax3.scatter(
                [plot_shift[0]],
                [plot_shift[1]],
                fc="white",
                ec="black",
                marker="X",
                s=100,
                zorder=10,
            )
        else:
            fig, ax = plt.subplots(
                figsize=(5, 5 * ratio),
                dpi=dpi,
            )
            self._plot_heatmap(
                fig=fig,
                ax=ax,
                X=X_plot,
                Y=Y_plot,
                Z=Z_both / (self.interface.area),
                borders=borders,
                cmap=cmap,
                fontsize=fontsize,
                show_max=show_max,
            )

            frac_shifts = np.c_[
                X_plot.ravel(), Y_plot.ravel(), np.zeros(Y_plot.shape).ravel()
            ].dot(np.linalg.inv(self.matrix))
            opt_shift = frac_shifts[np.argmax(Z_both.ravel())]
            max_Z = np.max(Z_both)
            plot_shift = opt_shift.dot(self.matrix)

            ax.scatter(
                [plot_shift[0]],
                [plot_shift[1]],
                fc="white",
                ec="black",
                marker="X",
                s=100,
                zorder=10,
            )

        self.opt_xy_shift = opt_shift

        fig.tight_layout()
        fig.savefig(output, bbox_inches="tight")
        plt.close(fig)

        return max_Z

    def run_z_shifts_old(
        self,
        interfacial_distances: np.ndarray = np.linspace(2.0, 4.0, 101),
        fontsize: int = 12,
        output: str = "z_shift.png",
        shift: bool = True,
    ) -> float:
        z_shifts = interfacial_distances - self.d_interface
        shifts = np.c_[
            np.zeros((len(z_shifts), 2)), z_shifts
        ] / np.linalg.norm(self.matrix[-1])
        inputs = self._generate_inputs(shifts)

        coulomb_energy = self._calculate_coulomb(inputs)
        born_energy = self._calculate_born(inputs)

        sub_coulomb_energy = coulomb_energy[0]
        film_coulomb_energy = coulomb_energy[1]
        interface_coulomb_energy = coulomb_energy[2:]

        sub_born_energy = born_energy[0]
        film_born_energy = born_energy[1]
        interface_born_energy = born_energy[2:]

        # pmg_coulomb_adh_energy = (
        #     sub_pmg_coulomb_energy + film_pmg_coulomb_energy
        # ) - interface_pmg_coulomb_energy
        coulomb_adh_energy = (
            sub_coulomb_energy + film_coulomb_energy
        ) - interface_coulomb_energy
        born_adh_energy = (
            sub_born_energy + film_born_energy
        ) - interface_born_energy

        both_adh_energy = born_adh_energy + coulomb_adh_energy
        # both_pmg_adh_energy = born_adh_energy + pmg_coulomb_adh_energy

        new_x = np.linspace(
            interfacial_distances.min(), interfacial_distances.max(), 500
        )
        new_z_shifts = np.linspace(
            shifts[:, -1].min(), shifts[:, -1].max(), 500
        )
        cs = CubicSpline(interfacial_distances, both_adh_energy)
        new_y = cs(new_x)

        max_ind = np.argmax(new_y)
        max_x = new_x[max_ind]
        max_y = new_y[max_ind]
        max_shift = new_z_shifts[max_ind]

        if shift:
            self.interface.shift_film(
                shift=[0, 0, max_shift], fractional=True, inplace=True
            )

        fig, ax = plt.subplots(figsize=(4, 4), dpi=400)
        ax.set_xlabel("Interfacial Distance ($\\AA$)", fontsize=fontsize)
        ax.set_ylabel("$E_{adh}$", fontsize=fontsize)
        ax.plot(new_x, new_y, color="black")
        ax.scatter([max_x], [max_y], color="black", marker="o", s=20)
        ax.set_xlim(interfacial_distances.min(), interfacial_distances.max())

        fig.tight_layout(pad=0.4)
        fig.savefig(output)
        plt.close(fig)

        return max_y

    def run_z_shifts(
        self,
        interfacial_distances: np.ndarray = np.linspace(2.0, 4.0, 101),
        fontsize: int = 12,
        output: str = "z_shift.png",
        shift: bool = True,
    ) -> float:
        z_shifts = interfacial_distances - self.d_interface
        shifts = np.c_[
            np.zeros((len(z_shifts), 2)), z_shifts
        ] / np.linalg.norm(self.matrix[-1])
        atoms_list = self._get_shifted_atoms_z(shifts)
        inputs = self._generate_inputs(atoms_list)

        coulomb_energy = self._calculate_coulomb(inputs)
        born_energy = self._calculate_born(inputs)

        sub_film_coulomb_energy = coulomb_energy[0]
        interface_coulomb_energy = coulomb_energy[1:]

        sub_film_born_energy = born_energy[0]
        interface_born_energy = born_energy[1:]

        coulomb_adh_energy = sub_film_coulomb_energy - interface_coulomb_energy
        born_adh_energy = sub_film_born_energy - interface_born_energy

        both_adh_energy = born_adh_energy + coulomb_adh_energy

        max_ind = np.argmax(both_adh_energy)
        max_x = interfacial_distances[max_ind]
        max_y = both_adh_energy[max_ind]
        max_shift = shifts[max_ind]

        self.opt_z_shift = max_shift

        fig, ax = plt.subplots(figsize=(4, 4), dpi=400)
        ax.set_xlabel("Interfacial Distance ($\\AA$)", fontsize=fontsize)
        ax.set_ylabel("$E_{adh}$", fontsize=fontsize)
        ax.plot(
            interfacial_distances,
            both_adh_energy,
            color="black",
        )
        ax.scatter([max_x], [max_y], color="black", marker="o", s=20)
        ax.set_xlim(interfacial_distances.min(), interfacial_distances.max())

        fig.tight_layout(pad=0.4)
        fig.savefig(output)
        plt.close(fig)

        return max_y



from OgreInterface.generate import SurfaceGenerator
from OgreInterface.lattice_match import ZurMcGill

from pymatgen.core.structure import Structure
from pymatgen.io.ase import AseAtomsAdaptor
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.core.surface import get_symmetrically_distinct_miller_indices

from ase import Atoms
import numpy as np

import matplotlib.pyplot as plt
from matplotlib.collections import PatchCollection
from matplotlib.colors import Normalize
from mpl_toolkits.axes_grid1.axes_divider import make_axes_locatable

from functools import reduce
from itertools import product

from typing import Union


class MillerSearch(object):
    """Class to perform a miller index scan to find all domain matched interfaces of various surfaces.

    Examples:
        >>> from OgreInterface.miller import MillerSearch
        >>> ms = MillerSearch(substrate="POSCAR_sub", film="POSCAR_film", max_substrate_index=1, max_film_index=1)
        >>> ms.run_scan()
        >>> ms.plot_misfits(output="miller_scan.png")

    Args:
        substrate: Bulk structure of the substrate in either Pymatgen Structure, ASE Atoms, or a structure file such as a POSCAR or Cif
        film: Bulk structure of the film in either Pymatgen Structure, ASE Atoms, or a structure file such as a POSCAR or Cif
        max_substrate_index: Max miller index of the substrate surfaces
        max_film_index: Max miller index of the film surfaces
        max_area_mismatch: Area ratio mismatch tolerance for the InterfaceGenerator
        max_angle_strain: Angle strain tolerance for the InterfaceGenerator
        max_linear_strain: Lattice vectors length mismatch tolerance for the InterfaceGenerator
        max_area: Maximum area of the matched supercells
        convert_to_conventional: Determines if the input structure is converted to it's conventional standard structure according to
            the Pymatgen SpacegroupAnalyzer.

    Attributes:
        substrate (Structure): Pymatgen Structure of the substrate
        film (Structure): Pymatgen Structure of the film
        max_substrate_index (int): Max miller index of the substrate surfaces
        max_film_index (int): Max miller index of the film surfaces
        max_area_mismatch (float): Area ratio mismatch tolerance for the InterfaceGenerator
        max_angle_strain (float): Angle strain tolerance for the InterfaceGenerator
        max_linear_strain (float): Lattice vectors length mismatch tolerance for the InterfaceGenerator
        max_area (float): Maximum area of the matched supercells
        convert_to_conventional (bool): Determines if the input structure is converted to it's conventional standard structure according to
            the Pymatgen SpacegroupAnalyzer.
        substrate_inds (list): List of unique substrate surface miller indices
        film_inds (list): List of unique film surface miller indices
    """

    def __init__(
        self,
        substrate: Union[Structure, Atoms, str],
        film: Union[Structure, Atoms, str],
        max_substrate_index: int = 1,
        max_film_index: int = 1,
        max_area_mismatch: float = 0.01,
        max_angle_strain: float = 0.01,
        max_linear_strain: float = 0.01,
        max_area: float = 500.0,
        convert_to_conventional: bool = True,
    ) -> None:
        self.convert_to_conventional = convert_to_conventional
        if type(substrate) == str:
            self.substrate, _ = self._get_bulk(Structure.from_file(substrate))
        else:
            self.substrate, _ = self._get_bulk(substrate)

        if type(film) == str:
            self.film, _ = self._get_bulk(Structure.from_file(film))
        else:
            self.film, _ = self._get_bulk(film)

        self.max_film_index = max_film_index
        self.max_substrate_index = max_substrate_index
        self.max_area_mismatch = max_area_mismatch
        self.max_angle_strain = max_angle_strain
        self.max_linear_strain = max_linear_strain
        self.max_area = max_area
        self.substrate_inds = self._get_unique_miller_indices(
            self.substrate, self.max_substrate_index
        )
        self.film_inds = self._get_unique_miller_indices(
            self.film, self.max_film_index
        )
        self._misfit_data = None
        self._area_data = None

    def _get_bulk(self, atoms_or_struc):
        if type(atoms_or_struc) == Atoms:
            init_structure = AseAtomsAdaptor.get_structure(atoms_or_struc)
        elif type(atoms_or_struc) == Structure:
            init_structure = atoms_or_struc
        else:
            raise TypeError(
                f"structure accepts 'pymatgen.core.structure.Structure' or 'ase.Atoms' not '{type(atoms_or_struc).__name__}'"
            )

        if self.convert_to_conventional:
            sg = SpacegroupAnalyzer(init_structure)
            conventional_structure = sg.get_conventional_standard_structure()
            conventional_atoms = AseAtomsAdaptor.get_atoms(
                conventional_structure
            )

            return conventional_structure, conventional_atoms
        else:
            init_atoms = AseAtomsAdaptor().get_atoms(init_structure)

            return init_structure, init_atoms

    def _float_gcd(self, a, b, rtol=1e-05, atol=1e-08):
        t = min(abs(a), abs(b))
        while abs(b) > rtol * t + atol:
            a, b = b, a % b
        return a

    def _hex_to_cubic(self, uvtw):
        u = 2 * uvtw[0] + uvtw[1]
        v = 2 * uvtw[1] + uvtw[0]
        w = uvtw[-1]

        output = np.array([u, v, w])
        gcd = np.abs(reduce(self._float_gcd, output))
        output = output / gcd

        return output.astype(int)

    def _cubic_to_hex(self, uvw):
        u = (1 / 3) * ((2 * uvw[0]) - uvw[1])
        v = (1 / 3) * ((2 * uvw[1]) - uvw[0])
        t = -(u + v)
        w = uvw[-1]

        output = np.array([u, v, t, w])
        gcd = np.abs(reduce(self._float_gcd, output))
        output /= gcd

        return output.astype(int)

    def _get_unique_miller_indices(self, struc, max_index):
        struc_sg = SpacegroupAnalyzer(struc)
        lattice = struc.lattice
        recip_lattice = lattice.reciprocal_lattice_crystallographic
        symmops = struc_sg.get_symmetry_operations(cartesian=False)
        recip_symmops = lattice.get_recp_symmetry_operation()
        planes = set(list(product(range(-max_index, max_index + 1), repeat=3)))
        planes.remove((0, 0, 0))

        reduced_planes = []
        for plane in planes:
            gcd = np.abs(reduce(self._float_gcd, plane))
            reduced_plane = tuple((plane / gcd).astype(int))
            reduced_planes.append(reduced_plane)

        reduced_planes = set(reduced_planes)

        planes_dict = {p: [] for p in reduced_planes}

        for plane in reduced_planes:
            if plane in planes_dict.keys():
                for i, symmop in enumerate(recip_symmops):
                    origin = symmop.operate((0, 0, 0))
                    point_out = symmop.operate(plane) - origin
                    gcd = np.abs(reduce(self._float_gcd, point_out))
                    point_out = tuple((point_out / gcd).astype(int))
                    planes_dict[plane].append(point_out)
                    if point_out != plane:
                        if point_out in planes_dict.keys():
                            del planes_dict[point_out]

        unique_planes = []

        for k in planes_dict:
            equivalent_planes = np.array(list(set(planes_dict[k])))
            diff = np.abs(np.sum(np.sign(equivalent_planes), axis=1))
            like_signs = equivalent_planes[diff == np.max(diff)]
            if len(like_signs) == 1:
                unique_planes.append(like_signs[0])
            else:
                first_max = like_signs[
                    np.abs(like_signs)[:, 0]
                    == np.max(np.abs(like_signs)[:, 0])
                ]
                if len(first_max) == 1:
                    unique_planes.append(first_max[0])
                else:
                    second_max = first_max[
                        np.abs(first_max)[:, 1]
                        == np.max(np.abs(first_max)[:, 1])
                    ]
                    if len(second_max) == 1:
                        unique_planes.append(second_max[0])
                    else:
                        unique_planes.append(
                            second_max[
                                np.argmax(np.sign(second_max).sum(axis=1))
                            ]
                        )

        unique_planes = np.vstack(unique_planes)
        sorted_planes = sorted(
            unique_planes, key=lambda x: (np.linalg.norm(x), -np.sign(x).sum())
        )

        return np.vstack(sorted_planes)

    def _get_unique_miller_indices_old(self, struc, max_index):
        struc_sg = SpacegroupAnalyzer(struc)
        # struc_conv_cell = struc_sg.get_conventional_standard_structure()
        symmops = struc_sg.get_symmetry_operations(cartesian=False)
        planes = set(list(product(range(-max_index, max_index + 1), repeat=3)))
        planes.remove((0, 0, 0))

        reduced_planes = []
        for plane in planes:
            gcd = np.abs(reduce(self._float_gcd, plane))
            reduced_plane = tuple((plane / gcd).astype(int))
            reduced_planes.append(reduced_plane)

        reduced_planes = set(reduced_planes)

        planes_dict = {p: [] for p in reduced_planes}

        for plane in reduced_planes:
            if plane in planes_dict.keys():
                for i, symmop in enumerate(symmops):
                    origin = symmop.operate((0, 0, 0))
                    point_out = symmop.operate(plane) - origin
                    gcd = np.abs(reduce(self._float_gcd, point_out))
                    point_out = tuple((point_out / gcd).astype(int))
                    planes_dict[plane].append(point_out)
                    if point_out != plane:
                        if point_out in planes_dict.keys():
                            del planes_dict[point_out]

        unique_planes = []

        for k in planes_dict:
            equivalent_planes = np.array(list(set(planes_dict[k])))
            diff = np.abs(np.sum(np.sign(equivalent_planes), axis=1))
            like_signs = equivalent_planes[diff == np.max(diff)]
            if len(like_signs) == 1:
                unique_planes.append(like_signs[0])
            else:
                first_max = like_signs[
                    np.abs(like_signs)[:, 0]
                    == np.max(np.abs(like_signs)[:, 0])
                ]
                if len(first_max) == 1:
                    unique_planes.append(first_max[0])
                else:
                    second_max = first_max[
                        np.abs(first_max)[:, 1]
                        == np.max(np.abs(first_max)[:, 1])
                    ]
                    if len(second_max) == 1:
                        unique_planes.append(second_max[0])
                    else:
                        unique_planes.append(
                            second_max[
                                np.argmax(np.sign(second_max).sum(axis=1))
                            ]
                        )

        unique_planes = np.vstack(unique_planes)
        sorted_planes = sorted(
            unique_planes, key=lambda x: (np.linalg.norm(x), -np.sign(x).sum())
        )

        return np.vstack(sorted_planes)

    def run_scan(self) -> None:
        """
        Run the miller index scan by looping through all combinations of unique surface miller indices
        for the substrate and film.
        """
        substrates = []
        films = []

        for inds in self.substrate_inds:
            sg_sub = SurfaceGenerator(
                bulk=self.substrate,
                miller_index=inds,
                layers=5,
                vacuum=10,
                generate_all=False,
                lazy=True,
                convert_to_conventional=self.convert_to_conventional,
            )
            sub_inplane_vectors = sg_sub.inplane_vectors
            sub_area = np.linalg.norm(
                np.cross(sub_inplane_vectors[0], sub_inplane_vectors[1])
            )
            substrates.append(
                [sub_inplane_vectors, sub_area, sg_sub.uvw_basis]
            )

        for inds in self.film_inds:
            sg_film = SurfaceGenerator(
                bulk=self.film,
                miller_index=inds,
                layers=5,
                vacuum=10,
                generate_all=False,
                lazy=True,
                convert_to_conventional=self.convert_to_conventional,
            )
            film_inplane_vectors = sg_film.inplane_vectors
            film_area = np.linalg.norm(
                np.cross(film_inplane_vectors[0], film_inplane_vectors[1])
            )

            films.append([film_inplane_vectors, film_area, sg_film.uvw_basis])

        misfits = np.ones((len(substrates), len(films))) * np.nan
        areas = np.ones((len(substrates), len(films))) * np.nan

        for i, substrate in enumerate(substrates):
            for j, film in enumerate(films):
                zm = ZurMcGill(
                    film_vectors=film[0],
                    substrate_vectors=substrate[0],
                    film_basis=film[2],
                    substrate_basis=substrate[2],
                    max_area=self.max_area,
                    max_linear_strain=self.max_linear_strain,
                    max_angle_strain=self.max_angle_strain,
                    max_area_mismatch=self.max_area_mismatch,
                )
                matches = zm.run()

                if len(matches) > 0:
                    min_area_match = matches[0]
                    area = min_area_match.area
                    strain = min_area_match.linear_strain
                    misfits[i, j] = strain.max()
                    areas[i, j] = area / np.sqrt(substrate[1] * film[1])

        self.misfits = np.round(misfits.T, 8)
        self.areas = areas.T

    def plot_misfits(
        self,
        cmap: str = "magma",
        dpi: int = 400,
        output: str = "misfit_plot.png",
        fontsize: float = 12.0,
        figure_scale: float = 1.0,
        labelrotation: float = -20.0,
        substrate_label: Union[str, None] = None,
        film_label: Union[str, None] = None,
        show_in_colab: bool = False,
    ) -> None:
        """
        Plot the results of the miller index scan.

        Args:
            cmap: color map (matplotlib)
            dpi: dpi (dots per inch) of the output image.
                Setting dpi=100 gives reasonably sized images when viewed in colab notebook
            output: File path for the output image
            fontsize: fontsize for axis and tick labels
            figure_scale: The figure size is automatically changed to fit the ratio of the substrate / film indices
                but in some cases, especially with large amounts of unique surfaces the figure size needs to be increased.
                This should usually stay at 1.0.
            labelrotation: Determines how much the labels on the x-axis should be rotated. This is usefull to avoid overlapping labels
            substrate_label: If none, this is automatically determined using the reduced formula of the bulk structure
            film_label: If none, this is automatically determined using the reduced formula of the bulk structure
            show_in_colab: Determines if the matplotlib figure is closed or not after the plot if made.
                if show_in_colab=True the plot will show up after you run the cell in colab/jupyter notebook.
        """
        ylabels = []
        for ylabel in self.film_inds:
            tmp_label = [
                str(i) if i >= 0 else "$\\overline{" + str(-i) + "}$"
                for i in ylabel
            ]
            ylabels.append(f'({"".join(tmp_label)})')

        xlabels = []
        for xlabel in self.substrate_inds:
            tmp_label = [
                str(i) if i >= 0 else "$\\overline{" + str(-i) + "}$"
                for i in xlabel
            ]
            xlabels.append(f'({"".join(tmp_label)})')

        N = len(self.film_inds)
        M = len(self.substrate_inds)
        x, y = np.meshgrid(np.arange(M), np.arange(N))
        s = self.areas
        c = self.misfits * 100

        if (M / N) < 1.0:
            figsize = (figure_scale * 5, (N / M) * figure_scale * 4)
        else:
            figsize = (figure_scale * 5 * (M / N), figure_scale * 4)

        fig, ax = plt.subplots(figsize=figsize, dpi=dpi)
        ax_divider = make_axes_locatable(ax)

        cax = ax_divider.append_axes(
            "right",
            size=np.min(figsize) * 0.04,
            pad=np.min(figsize) * 0.01,
        )

        if film_label is None:
            film_label = self.film.composition.reduced_formula

        ax.set_ylabel(film_label + " Miller Index", fontsize=fontsize)

        if substrate_label is None:
            substrate_label = self.substrate.composition.reduced_formula

        ax.set_xlabel(substrate_label + " Miller Index", fontsize=fontsize)

        R = 0.85 * s / np.nanmax(s) / 2
        circles = [
            plt.Circle((i, j), radius=r, edgecolor="black", lw=3)
            for r, i, j in zip(R.flat, x.flat, y.flat)
        ]
        col = PatchCollection(
            circles,
            array=c.flatten(),
            cmap=cmap,
            norm=Normalize(
                vmin=np.max([np.nanmin(c) - 0.01 * np.nanmin(c), 0]),
                vmax=np.nanmax(c) + 0.01 * np.nanmax(c),
            ),
            edgecolor="black",
            linewidth=1,
        )
        ax.add_collection(col)

        ax.set(
            xticks=np.arange(M),
            yticks=np.arange(N),
            xticklabels=xlabels,
            yticklabels=ylabels,
        )
        ax.set_xticks(np.arange(M + 1) - 0.5, minor=True)
        ax.set_yticks(np.arange(N + 1) - 0.5, minor=True)
        ax.tick_params(axis="x", labelrotation=labelrotation)
        ax.tick_params(labelsize=fontsize)
        ax.grid(which="minor", linestyle=":", linewidth=0.75)

        cbar = fig.colorbar(col, cax=cax)
        cbar.set_label("Misfit Percentage", fontsize=fontsize)
        cbar.ax.tick_params(labelsize=fontsize)
        cbar.ax.ticklabel_format(
            style="sci", scilimits=(-3, 3), useMathText=True
        )
        cbar.ax.yaxis.set_offset_position("left")

        ax.set_aspect("equal")
        fig.tight_layout(pad=0.4)
        fig.savefig(output, bbox_inches="tight")

        if not show_in_colab:
            plt.close(fig)


if __name__ == "__main__":
    ms = MillerSearch(
        substrate="./dd-poscars/POSCAR_InAs_conv",
        film="./dd-poscars/POSCAR_Al_conv",
        max_film_index=2,
        max_substrate_index=2,
        max_linear_strain=0.01,
        max_angle_strain=0.01,
        max_area_mismatch=0.01,
        max_area=500,
    )
    ms.run_scan()
    ms.plot_misfits(figsize=(6.5, 5), fontsize=17, labelrotation=0)



from __future__ import annotations
import numpy as np
import copy
from functools import reduce
from pymatgen.core.structure import Structure
from pymatgen.core.lattice import Lattice
import itertools
import functools
import math
import collections


def conv_a_to_b(struc_a: Structure, struc_b: Structure) -> np.ndarray:
    return struc_b.lattice.matrix @ struc_a.lattice.inv_matrix


def get_layer_supercelll(
    structure: Structure, layers: int, vacuum_scale: int = 0
) -> Structure:
    base_frac_coords = structure.frac_coords
    sc_base_frac_coords = np.vstack(
        [base_frac_coords + np.array([0, 0, i]) for i in range(layers)]
    )
    sc_cart_coords = sc_base_frac_coords.dot(structure.lattice.matrix)
    sc_layer_inds = np.repeat(np.arange(layers), len(structure))

    new_site_properties = {
        k: v * layers for k, v in structure.site_properties.items()
    }
    new_site_properties["layer_index"] = sc_layer_inds.tolist()

    layer_transform = np.eye(3)
    layer_transform[-1, -1] = layers + vacuum_scale
    layer_matrix = layer_transform @ structure.lattice.matrix

    layer_slab = Structure(
        lattice=Lattice(matrix=layer_matrix),
        species=structure.species * layers,
        coords=sc_cart_coords,
        coords_are_cartesian=True,
        to_unit_cell=True,
        site_properties=new_site_properties,
    )

    return layer_slab


def group_layers(structure, atol=None):
    """
    This function will find the atom indices belonging to each unique atomic layer.

    Parameters:
        structure (pymatgen.core.structure.Structure): Slab structure
        atol (float or None): Tolarence used for grouping the layers. Useful for grouping
            layers in a structure with relaxed atomic positions.

    Returns:
        A list containing the indices of each layers.
        A list of heights of each layers in fractional coordinates.
    """
    sites = structure.sites
    zvals = np.array([site.c for site in sites])
    unique_values = np.sort(np.unique(np.round(zvals, 3)))
    diff = np.mean(np.diff(unique_values)) * 0.2

    grouped = False
    groups = []
    group_heights = []
    zvals_copy = copy.deepcopy(zvals)
    while not grouped:
        if len(zvals_copy) > 0:
            if atol is None:
                group_index = np.where(
                    np.isclose(zvals, np.min(zvals_copy), atol=diff)
                )[0]
            else:
                group_index = np.where(
                    np.isclose(zvals, np.min(zvals_copy), atol=atol)
                )[0]

            group_heights.append(np.min(zvals_copy))
            zvals_copy = np.delete(
                zvals_copy,
                np.where(np.isin(zvals_copy, zvals[group_index]))[0],
            )
            groups.append(group_index)
        else:
            grouped = True

    return groups, np.array(group_heights)


def get_reduced_basis(basis: np.ndarray) -> np.ndarray:
    """
    This function is used to find the miller indices of the slab structure
    basis vectors in their most reduced form. i.e.

    |  2  4  0 |     | 1  2  0 |
    |  0 -2  4 | ==> | 0 -1  2 |
    | 10 10 10 |     | 1  1  1 |

    Parameters:
        basis (np.ndarray): 3x3 matrix defining the lattice vectors

    Returns:
        Reduced integer basis in the form of miller indices
    """
    basis /= np.linalg.norm(basis, axis=1)[:, None]

    for i, b in enumerate(basis):
        basis[i] = _get_reduced_vector(b)

    return basis


def _get_reduced_vector(vector: np.ndarry) -> np.ndarray:
    """ """
    abs_b = np.abs(vector)
    vector /= abs_b[abs_b > 0.001].min()
    vector /= np.abs(reduce(_float_gcd, vector))

    return vector


def _float_gcd(a, b, rtol=1e-05, atol=1e-08):
    t = min(abs(a), abs(b))
    while abs(b) > rtol * t + atol:
        a, b = b, a % b
    return a


def reduce_vectors_zur_and_mcgill(a, b):
    vecs = np.vstack([a, b])
    mat = np.eye(3)
    reduced = False

    while not reduced:
        if np.dot(vecs[0], vecs[1]) < 0:
            vecs[1] *= -1
            mat[1] *= -1
            continue

        if np.linalg.norm(vecs[0]) > np.linalg.norm(vecs[1]):
            vecs = vecs[[1, 0]]
            mat = mat[[1, 0, 2]]
            continue

        if np.linalg.norm(vecs[1]) > np.linalg.norm(vecs[1] + vecs[0]):
            vecs[1] = vecs[1] + vecs[0]
            mat[1] = mat[1] + mat[0]
            continue

        if np.linalg.norm(vecs[1]) > np.linalg.norm(vecs[1] - vecs[0]):
            vecs[1] = vecs[1] - vecs[0]
            mat[1] = mat[1] - mat[0]
            reduced = True
            continue

        reduced = True

    return vecs[0], vecs[1], mat


def get_primitive_structure(
    struc,
    tolerance: float = 0.25,
    use_site_props: bool = False,
    constrain_latt: list | dict | None = None,
):
    """
    This finds a smaller unit cell than the input. Sometimes it doesn"t
    find the smallest possible one, so this method is recursively called
    until it is unable to find a smaller cell.

    NOTE: if the tolerance is greater than 1/2 the minimum inter-site
    distance in the primitive cell, the algorithm will reject this lattice.

    Args:
        tolerance (float), Angstroms: Tolerance for each coordinate of a
            particular site. For example, [0.1, 0, 0.1] in cartesian
            coordinates will be considered to be on the same coordinates
            as [0, 0, 0] for a tolerance of 0.25. Defaults to 0.25.
        use_site_props (bool): Whether to account for site properties in
            differentiating sites.
        constrain_latt (list/dict): List of lattice parameters we want to
            preserve, e.g. ["alpha", "c"] or dict with the lattice
            parameter names as keys and values we want the parameters to
            be e.g. {"alpha": 90, "c": 2.5}.

    Returns:
        The most primitive structure found.
    """
    if constrain_latt is None:
        constrain_latt = []

    def site_label(site):
        if not use_site_props:
            return site.species_string
        d = [site.species_string]
        for k in sorted(site.properties.keys()):
            d.append(k + "=" + str(site.properties[k]))
        return ", ".join(d)

    # group sites by species string
    sites = sorted(struc._sites, key=site_label)

    grouped_sites = [
        list(a[1]) for a in itertools.groupby(sites, key=site_label)
    ]
    grouped_fcoords = [
        np.array([s.frac_coords for s in g]) for g in grouped_sites
    ]

    # min_vecs are approximate periodicities of the cell. The exact
    # periodicities from the supercell matrices are checked against these
    # first
    min_fcoords = min(grouped_fcoords, key=lambda x: len(x))
    min_vecs = min_fcoords - min_fcoords[0]

    # fractional tolerance in the supercell
    super_ftol = np.divide(tolerance, struc.lattice.abc)
    super_ftol_2 = super_ftol * 2

    def pbc_coord_intersection(fc1, fc2, tol):
        """
        Returns the fractional coords in fc1 that have coordinates
        within tolerance to some coordinate in fc2
        """
        d = fc1[:, None, :] - fc2[None, :, :]
        d -= np.round(d)
        np.abs(d, d)
        return fc1[np.any(np.all(d < tol, axis=-1), axis=-1)]

    # here we reduce the number of min_vecs by enforcing that every
    # vector in min_vecs approximately maps each site onto a similar site.
    # The subsequent processing is O(fu^3 * min_vecs) = O(n^4) if we do no
    # reduction.
    # This reduction is O(n^3) so usually is an improvement. Using double
    # the tolerance because both vectors are approximate
    for g in sorted(grouped_fcoords, key=lambda x: len(x)):
        for f in g:
            min_vecs = pbc_coord_intersection(min_vecs, g - f, super_ftol_2)

    def get_hnf(fu):
        """
        Returns all possible distinct supercell matrices given a
        number of formula units in the supercell. Batches the matrices
        by the values in the diagonal (for less numpy overhead).
        Computational complexity is O(n^3), and difficult to improve.
        Might be able to do something smart with checking combinations of a
        and b first, though unlikely to reduce to O(n^2).
        """

        def factors(n):
            for i in range(1, n + 1):
                if n % i == 0:
                    yield i

        for det in factors(fu):
            if det == 1:
                continue
            for a in factors(det):
                for e in factors(det // a):
                    g = det // a // e
                    yield det, np.array(
                        [
                            [[a, b, c], [0, e, f], [0, 0, g]]
                            for b, c, f in itertools.product(
                                range(a), range(a), range(e)
                            )
                        ]
                    )

    # we can't let sites match to their neighbors in the supercell
    grouped_non_nbrs = []
    for gfcoords in grouped_fcoords:
        fdist = gfcoords[None, :, :] - gfcoords[:, None, :]
        fdist -= np.round(fdist)
        np.abs(fdist, fdist)
        non_nbrs = np.any(fdist > 2 * super_ftol[None, None, :], axis=-1)
        # since we want sites to match to themselves
        np.fill_diagonal(non_nbrs, True)
        grouped_non_nbrs.append(non_nbrs)

    num_fu = functools.reduce(math.gcd, map(len, grouped_sites))
    for size, ms in get_hnf(num_fu):
        inv_ms = np.linalg.inv(ms)

        # find sets of lattice vectors that are are present in min_vecs
        dist = inv_ms[:, :, None, :] - min_vecs[None, None, :, :]
        dist -= np.round(dist)
        np.abs(dist, dist)
        is_close = np.all(dist < super_ftol, axis=-1)
        any_close = np.any(is_close, axis=-1)
        inds = np.all(any_close, axis=-1)

        for inv_m, m in zip(inv_ms[inds], ms[inds]):
            new_m = np.dot(inv_m, struc.lattice.matrix)
            ftol = np.divide(tolerance, np.sqrt(np.sum(new_m**2, axis=1)))

            valid = True
            new_coords = []
            new_sp = []
            new_props = collections.defaultdict(list)
            for gsites, gfcoords, non_nbrs in zip(
                grouped_sites, grouped_fcoords, grouped_non_nbrs
            ):
                all_frac = np.dot(gfcoords, m)

                # calculate grouping of equivalent sites, represented by
                # adjacency matrix
                fdist = all_frac[None, :, :] - all_frac[:, None, :]
                fdist = np.abs(fdist - np.round(fdist))
                close_in_prim = np.all(fdist < ftol[None, None, :], axis=-1)
                groups = np.logical_and(close_in_prim, non_nbrs)

                # check that groups are correct
                if not np.all(np.sum(groups, axis=0) == size):
                    valid = False
                    break

                # check that groups are all cliques
                for g in groups:
                    if not np.all(groups[g][:, g]):
                        valid = False
                        break
                if not valid:
                    break

                # add the new sites, averaging positions
                added = np.zeros(len(gsites))
                new_fcoords = all_frac % 1
                for i, group in enumerate(groups):
                    if not added[i]:
                        added[group] = True
                        inds = np.where(group)[0]
                        coords = new_fcoords[inds[0]]
                        for n, j in enumerate(inds[1:]):
                            offset = new_fcoords[j] - coords
                            coords += (offset - np.round(offset)) / (n + 2)
                        new_sp.append(gsites[inds[0]].species)
                        for k in gsites[inds[0]].properties:
                            new_props[k].append(gsites[inds[0]].properties[k])
                        new_coords.append(coords)

            if valid:
                inv_m = np.linalg.inv(m)
                new_l = Lattice(np.dot(inv_m, struc.lattice.matrix))
                s = Structure(
                    new_l,
                    new_sp,
                    new_coords,
                    site_properties=new_props,
                    coords_are_cartesian=False,
                )

                p = get_primitive_structure(
                    s,
                    tolerance=tolerance,
                    use_site_props=use_site_props,
                    constrain_latt=constrain_latt,
                )
                if not constrain_latt:
                    return p

                # Only return primitive structures that
                # satisfy the restriction condition
                p_latt, s_latt = p.lattice, struc.lattice
                if type(constrain_latt).__name__ == "list":
                    if all(
                        getattr(p_latt, pp) == getattr(s_latt, pp)
                        for pp in constrain_latt
                    ):
                        return p
                elif type(constrain_latt).__name__ == "dict":
                    if all(
                        getattr(p_latt, pp) == constrain_latt[pp] for pp in constrain_latt.keys()  # type: ignore
                    ):
                        return p

    return struc.copy()



from OgreInterface.score_function.scatter import scatter_add
from typing import Dict
import numpy as np
import torch.nn as nn
import torch


class EnergyEwaldError(Exception):
    pass


class EnergyEwald(nn.Module):
    """
    Compute the Coulomb energy of a set of point charges inside a periodic box.
    Only works for periodic boundary conditions in all three spatial directions and orthorhombic boxes.
    Args:
        alpha (float): Ewald alpha.
        k_max (int): Number of lattice vectors.
        charges_key (str): Key of partial charges in the input batch.
    """

    def __init__(
        self,
        alpha: float,
        k_max: int,
    ):
        super(EnergyEwald, self).__init__()

        # Get the appropriate Coulomb constant
        ke = 14.3996
        self.register_buffer("ke", torch.Tensor([ke]))

        # TODO: automatic computation of alpha
        self.register_buffer("alpha", torch.Tensor([alpha]))

        # Set up lattice vectors
        self.k_max = k_max
        kvecs = self._generate_kvecs()
        self.register_buffer("kvecs", kvecs)

    def _generate_kvecs(self) -> torch.Tensor:
        """
        Auxiliary routine for setting up the k-vectors.
        Returns:
            torch.Tensor: k-vectors.
        """
        krange = torch.arange(0, self.k_max + 1, dtype=self.alpha.dtype)
        krange = torch.cat([krange, -krange[1:]])
        kvecs = torch.cartesian_prod(krange, krange, krange)
        norm = torch.sum(kvecs**2, dim=1)
        kvecs = kvecs[norm <= self.k_max**2 + 2, :]
        norm = norm[norm <= self.k_max**2 + 2]
        kvecs = kvecs[norm != 0, :]

        return kvecs

    def forward(
        self, inputs: Dict[str, torch.Tensor]
    ) -> Dict[str, torch.Tensor]:
        """
        Compute the Coulomb energy of the periodic system.
        Args:
            inputs (dict(str,torch.Tensor)): Input batch.
        Returns:
            dict(str, torch.Tensor): results with Coulomb energy.
        """
        q = inputs["partial_charges"].squeeze(-1)
        idx_m = inputs["idx_m"]

        r_ij = inputs["Rij"]
        idx_i = inputs["idx_i"]
        idx_j = inputs["idx_j"]

        d_ij = torch.norm(r_ij, dim=1)

        positions = inputs["R"]
        cell = inputs["cell"]

        n_atoms = q.shape[0]
        n_molecules = int(idx_m[-1]) + 1

        # Get real space and reciprocal space contributions
        y_real = self._real_space(
            q, d_ij, idx_i, idx_j, idx_m, n_atoms, n_molecules
        )
        y_reciprocal = self._reciprocal_space(
            q, positions, cell, idx_m, n_molecules
        )

        y = y_real + y_reciprocal

        return y.numpy()

    def _real_space(
        self,
        q: torch.Tensor,
        d_ij: torch.Tensor,
        idx_i: torch.Tensor,
        idx_j: torch.Tensor,
        idx_m: torch.Tensor,
        n_atoms: int,
        n_molecules: int,
    ) -> torch.Tensor:
        """
        Compute the real space contribution of the screened charges.
        Args:
            q (torch.Tensor): Partial charges.
            d_ij (torch.Tensor): Interatomic distances.
            idx_i (torch.Tensor): Indices of atoms i in the distance pairs.
            idx_j (torch.Tensor): Indices of atoms j in the distance pairs.
            idx_m (torch.Tensor): Molecular indices of each atom.
            n_atoms (int): Total number of atoms.
            n_molecules (int): Number of molecules.
        Returns:
            torch.Tensor: Real space Coulomb energy.
        """

        # Apply erfc for Ewald summation
        f_erfc = torch.erfc(torch.sqrt(self.alpha) * d_ij)
        # Combine functions and multiply with inverse distance
        f_r = f_erfc / d_ij

        potential_ij = q[idx_i] * q[idx_j] * f_r

        y = scatter_add(potential_ij, idx_i, dim_size=n_atoms)
        y = scatter_add(y, idx_m, dim_size=n_molecules)
        y = 0.5 * self.ke * y.squeeze(-1)

        return y

    def _reciprocal_space(
        self,
        q: torch.Tensor,
        positions: torch.Tensor,
        cell: torch.Tensor,
        idx_m: torch.Tensor,
        n_molecules: int,
    ):
        """
        Compute the reciprocal space contribution.
        Args:
            q (torch.Tensor): Partial charges.
            positions (torch.Tensor): Atom positions.
            cell (torch.Tensor): Molecular cells.
            idx_m (torch.Tensor): Molecular indices of each atom.
            n_molecules (int): Number of molecules.
        Returns:
            torch.Tensor: Real space Coulomb energy.
        """
        # extract box dimensions from cells
        recip_box = 2.0 * np.pi * torch.linalg.inv(cell).transpose(1, 2)
        v_box = torch.abs(torch.linalg.det(cell))

        if torch.any(torch.isclose(v_box, torch.zeros_like(v_box))):
            raise EnergyEwaldError("Simulation box has no volume.")

        # 1) compute the prefactor
        prefactor = 2.0 * np.pi / v_box

        # setup kvecs M x K x 3
        kvecs = torch.matmul(self.kvecs[None, :, :], recip_box)

        # Squared length of vectors M x K
        k_squared = torch.sum(kvecs**2, dim=2)

        # 2) Gaussian part of ewald sum
        q_gauss = torch.exp(-0.25 * k_squared / self.alpha)  # M x K

        # 3) Compute charge density fourier terms
        # Dot product in exponent -> MN x K, expand kvecs in MN batch structure
        kvec_dot_pos = torch.sum(kvecs[idx_m] * positions[:, None, :], dim=2)

        # charge densities MN x K -> M x K
        q_real = scatter_add(
            (q[:, None] * torch.cos(kvec_dot_pos)), idx_m, dim_size=n_molecules
        )
        q_imag = scatter_add(
            (q[:, None] * torch.sin(kvec_dot_pos)), idx_m, dim_size=n_molecules
        )
        # Compute square of density
        q_dens = q_real**2 + q_imag**2

        # Sum over k vectors -> M x K -> M
        y_ewald = prefactor * torch.sum(q_dens * q_gauss / k_squared, dim=1)

        # 4) self interaction correction -> MN
        self_interaction = torch.sqrt(self.alpha / np.pi) * scatter_add(
            q**2, idx_m, dim_size=n_molecules
        )

        # Bring everything together
        y_ewald = self.ke * (y_ewald - self_interaction)

        return y_ewald


if __name__ == "__main__":
    from generate_inputs import generate_dict
    from ase.io import read

    ew = EnergyEwald(alpha, k_max)
    ew.forward(generate_dict(atoms, cutoff))



import torch
from torch import nn


def scatter_add(
    x: torch.Tensor, idx_i: torch.Tensor, dim_size: int, dim: int = 0
) -> torch.Tensor:
    """
    Sum over values with the same indices.
    Args:
        x: input values
        idx_i: index of center atom i
        dim_size: size of the dimension after reduction
        dim: the dimension to reduce
    Returns:
        reduced input
    """
    return _scatter_add(x, idx_i, dim_size, dim)


@torch.jit.script
def _scatter_add(
    x: torch.Tensor, idx_i: torch.Tensor, dim_size: int, dim: int = 0
) -> torch.Tensor:
    shape = list(x.shape)
    shape[dim] = dim_size
    tmp = torch.zeros(shape, dtype=x.dtype, device=x.device)
    y = tmp.index_add(dim, idx_i, x)
    return y



from ase.neighborlist import neighbor_list
from typing import Dict, List, Optional
from ase import Atoms
import torch
from OgreInterface.score_function.neighbors import TorchNeighborList


def _atoms_collate_fn(batch):
    """
    Build batch from systems and properties & apply padding
    Args:
        examples (list):
    Returns:
        dict[str->torch.Tensor]: mini-batch of atomistic systems
    """
    elem = batch[0]
    idx_keys = {"idx_i", "idx_j", "idx_i_triples"}
    # Atom triple indices must be treated separately
    idx_triple_keys = {"idx_j_triples", "idx_k_triples"}

    coll_batch = {}
    for key in elem:
        if (key not in idx_keys) and (key not in idx_triple_keys):
            coll_batch[key] = torch.cat([d[key] for d in batch], 0)
        elif key in idx_keys:
            coll_batch[key + "_local"] = torch.cat([d[key] for d in batch], 0)

    seg_m = torch.cumsum(coll_batch["n_atoms"], dim=0)
    seg_m = torch.cat([torch.zeros((1,), dtype=seg_m.dtype), seg_m], dim=0)
    idx_m = torch.repeat_interleave(
        torch.arange(len(batch)), repeats=coll_batch["n_atoms"], dim=0
    )
    coll_batch["idx_m"] = idx_m

    for key in idx_keys:
        if key in elem.keys():
            coll_batch[key] = torch.cat(
                [d[key] + off for d, off in zip(batch, seg_m)], 0
            )

    # Shift the indices for the atom triples
    for key in idx_triple_keys:
        if key in elem.keys():
            indices = []
            offset = 0
            for idx, d in enumerate(batch):
                indices.append(d[key] + offset)
                offset += d["idx_j"].shape[0]
            coll_batch[key] = torch.cat(indices, 0)

    return coll_batch


def generate_dict_ase(
    atoms: List[Atoms],
    cutoff: float,
    charge_dict: Dict[str, float],
    radius_dict: Dict[str, float],
    ns_dict: Dict[str, float],
) -> Dict:
    inputs_batch = []

    for at_idx, atom in enumerate(atoms):
        charges = torch.Tensor(
            [charge_dict[s] for s in atom.get_chemical_symbols()]
        )
        r0s = torch.Tensor(
            [radius_dict[s] for s in atom.get_chemical_symbols()]
        )
        is_film = torch.from_numpy(atom.get_array("is_film", copy=True))
        ns = torch.Tensor([ns_dict[s] for s in atom.get_chemical_symbols()])
        R = torch.from_numpy(atom.get_positions())
        cell = torch.from_numpy(atom.get_cell().array).view(-1, 3, 3)
        idx_i, idx_j, S = neighbor_list(
            "ijS", atom, cutoff, self_interaction=False
        )
        idx_i = torch.from_numpy(idx_i)
        idx_j = torch.from_numpy(idx_j)
        S = torch.from_numpy(S).to(dtype=R.dtype)
        offsets = torch.mm(S, cell.view(3, 3))
        Rij = R[idx_j] - R[idx_i] + offsets

        input_dict = {
            "n_atoms": torch.tensor([atom.get_global_number_of_atoms()]),
            "Z": torch.from_numpy(atom.get_atomic_numbers()),
            "R": R,
            "cell": cell,
            "idx_i": idx_i,
            "idx_j": idx_j,
            "Rij": Rij,
            "pbc": torch.from_numpy(atom.get_pbc()).view(-1, 3),
            "partial_charges": charges,
            "r0s": r0s,
            "ns": ns,
            "is_film": is_film,
        }

        inputs_batch.append(input_dict)

    inputs = _atoms_collate_fn(inputs_batch)

    for k, v in inputs.items():
        if "float" in str(v.dtype):
            inputs[k] = v.to(dtype=torch.float32)
        if "idx" in k:
            inputs[k] = v.to(dtype=torch.long)

    return inputs


def generate_dict_torch(
    atoms: List[Atoms],
    cutoff: float,
    ns_dict: Optional[Dict[str, float]] = None,
    charge_dict: Optional[Dict[str, float]] = None,
) -> Dict:

    tn = TorchNeighborList(cutoff=cutoff)
    inputs_batch = []

    for at_idx, atom in enumerate(atoms):

        is_film = torch.from_numpy(
            atom.get_array("is_film", copy=True).astype(int)
        )
        R = torch.from_numpy(atom.get_positions())
        cell = torch.from_numpy(atom.get_cell().array)

        input_dict = {
            "n_atoms": torch.tensor([atom.get_global_number_of_atoms()]),
            "Z": torch.from_numpy(atom.get_atomic_numbers()),
            "R": R,
            "cell": cell,
            "pbc": torch.from_numpy(atom.get_pbc()),
            "is_film": is_film,
        }

        if charge_dict is not None:
            charges = torch.Tensor(
                [charge_dict[s] for s in atom.get_chemical_symbols()]
            )
            ns = torch.Tensor(
                [ns_dict[s] for s in atom.get_chemical_symbols()]
            )
            input_dict["partial_charges"] = charges
            input_dict["ns"] = ns

        tn.forward(inputs=input_dict)

        Rij = (
            R[input_dict["idx_j"]]
            - R[input_dict["idx_i"]]
            + input_dict["offsets"]
        )

        input_dict["Rij"] = Rij
        input_dict["cell"] = input_dict["cell"].view(-1, 3, 3)
        input_dict["pbc"] = input_dict["pbc"].view(-1, 3)

        inputs_batch.append(input_dict)

    inputs = _atoms_collate_fn(inputs_batch)

    for k, v in inputs.items():
        if "float" in str(v.dtype):
            inputs[k] = v.to(dtype=torch.float32)
        if "idx" in k:
            inputs[k] = v.to(dtype=torch.long)

    return inputs


if __name__ == "__main__":
    from ase.build import bulk

    InAs = bulk("InAs", crystalstructure="zincblende", a=5.6)
    charge_dict = {"In": 0.0, "As": 0.0}
    inputs = generate_dict_torch([InAs], cutoff=10.0, charge_dict=charge_dict)
    print(inputs["n_atoms"])
